 (* my heater at home *)

open Data

type button = Up | Down

let static low = 4
let static high = 4

let static delay_on = 500 (* in miliseconds *)
let static delay_off = 100

(* [count d t] returns [true] when [d] occurrences of [t] has been received*)
let node count d t = ok where
  rec ok = cpt = 0
  and cpt = d -> if t & not (pre ok) then pre cpt - 1 else pre cpt

(* [edge x] returns the rising edge *)
let node edge x = false -> not (pre x) & x

(* controling the heat *)
(* returns [true] when [expected_temp] does not agree with [actual_temp] *)
(* we introduce an hysteresis to avoir oscilation *)
let node heat expected_temp actual_temp = ok where
  rec ok = if actual_temp <= expected_temp - low then true
           else if actual_temp >= expected_temp + high then false
           else false -> pre ok

(* the same function using two modes *)
let node heat expected_temp actual_temp = ok where
  rec automaton
	False -> 
	  do ok = false 
	  unless (actual_temp <= expected_temp - low) then True
      | True ->
	  do ok = true 
	  unless (actual_temp >= expected_temp + high) then False
      end

(* a cyclic two mode automaton with an internal timer *)
(* [open_light = true] and [open_gas = true] for [delay_on milisecond] *)
(* then [open_light = false] and [open_gas = false] for *)
(* [delay_off milisecond] *)
let node command milisecond = (open_light, open_gas) where
  rec automaton
	Open ->
	  do open_light = true
	  and open_gas = true
	  until (count delay_on milisecond) then Silent
      | Silent ->
	  do open_light = false
	  and open_gas = false
	  until (count delay_off milisecond) then Open
      end


(* the main command which control the opening of the light and gas *)
(* [on_heat] indicates that the water must be heated *)
(* at most three test must be made. At the end, if [on_light] is false *)
(* the system must be blocked into a security stop *)
let node light milisecond on_heat on_light = (open_light, open_gas, nok) 
where
  rec automaton
	Off_light ->
          do nok = false
	  and open_light = false
	  and open_gas = false
	  until on_heat then Try
      | On_light ->
          do nok = false
          and open_light = false
          and open_gas = true
          until (not on_heat) then Off_light
      | Try ->
	  do
	    (open_light, open_gas) = command milisecond
	  until on_light then On_light
          until (count 3 (edge (not open_light))) then Failure
      | Failure ->
	  do nok = true 
	  and open_light = false
	  and open_gas = false
	  done
      end

(* the main function *)
let node main milisecond restart expected_temp actual_temp on_light =
                                       (open_light, open_gas, ok, nok) where
  rec reset
	on_heat = heat expected_temp actual_temp
      and
        (open_light, open_gas,nok) = light milisecond on_heat on_light
      and
        ok = not nok       
      every restart

(** now add some stub code to prepare the simulation *)
(* we do something very naive here: the environment is *)
(* given by the user. In a more realistic simulation, we *)
(* should write a program for the heater itself (+ random) *)

(* transforming a double button into an integer value *)
let node button_to_integer init step button = temp where
  rec last temp = init
  and present
        button(v) ->
	  do match v with
	       Up -> do temp = last temp + step done
	     | Down -> do temp = last temp - step done
	     end
	  done
      end

(* the interface is done through the keybord *)
let node environment key =
  let 
    (* build the various input/output events from the environment *)
    match key with
      None -> do done
    | Some(v) ->
	do match v with
	     'P' -> do emit expected_temp = Up done
	   | 'M' -> do emit expected_temp = Down done
	   | 'E' -> do emit actual_temp = Up done
	   | 'Z' -> do emit actual_temp = Down done
	   | 'R' -> do emit restart = () done
	   | 'L' -> do emit on_light = () done
	   | _ -> do done
	   end
	done
    end in
  let
      expected_temp = button_to_integer 20 1 expected_temp in
  let
      actual_temp = button_to_integer 20 1 actual_temp in
  let
      on_light = ?(on_light) in
  let 
      restart = ?(restart) in

  expected_temp, actual_temp, on_light, restart

(* print values *)
let output open_light open_gas ok nok expected_temp actual_temp =
  print_string "open_light = ";
  print_string (if open_light then "on" else "off");
  print_string " open_gas = ";
  print_string (if open_gas then "on" else "off");
  print_string (if ok then " ok" else if nok then " nok" else "");
  print_string " expected_temp = ";
  print_int expected_temp;
  print_string " actual_temp = ";
  print_int actual_temp;
  print_newline ()

(* the base clock should be the milisecond *)
let node main () =
  let key = keyboard () in
  (* the external environment is simulated by the user *)
  (* only make sense for simulation purposes *)
  let expected_temp, actual_temp, on_light, restart = environment key in
  
  (* add the controler *)
  let open_light, open_gas, ok, nok = 
    main true restart expected_temp actual_temp on_light in
  
  (* output the values computed by the controler *)
  output open_light open_gas ok nok expected_temp actual_temp

